"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const react_1 = require("react");
function cursorPositionToTextOffset(cursorPos, body) {
    if (!body)
        return 0;
    const noteLines = body.split('\n');
    let pos = 0;
    for (let i = 0; i < noteLines.length; i++) {
        if (i > 0)
            pos++; // Need to add the newline that's been removed in the split() call above
        if (i === cursorPos.row) {
            pos += cursorPos.column;
            break;
        }
        else {
            pos += noteLines[i].length;
        }
    }
    return pos;
}
exports.cursorPositionToTextOffset = cursorPositionToTextOffset;
function currentTextOffset(editor, body) {
    return cursorPositionToTextOffset(editor.getCursorPosition(), body);
}
exports.currentTextOffset = currentTextOffset;
function rangeToTextOffsets(range, body) {
    return {
        start: cursorPositionToTextOffset(range.start, body),
        end: cursorPositionToTextOffset(range.end, body),
    };
}
exports.rangeToTextOffsets = rangeToTextOffsets;
function textOffsetSelection(selectionRange, body) {
    return selectionRange && body ? rangeToTextOffsets(selectionRange, body) : null;
}
exports.textOffsetSelection = textOffsetSelection;
function selectedText(selectionRange, body) {
    const selection = textOffsetSelection(selectionRange, body);
    if (!selection || selection.start === selection.end)
        return '';
    return body.substr(selection.start, selection.end - selection.start);
}
exports.selectedText = selectedText;
function selectionRange(editor) {
    const ranges = editor.getSelection().getAllRanges();
    return ranges && ranges.length ? ranges[0] : null;
}
exports.selectionRange = selectionRange;
function textOffsetToCursorPosition(offset, body) {
    const lines = body.split('\n');
    let row = 0;
    let currentOffset = 0;
    for (let i = 0; i < lines.length; i++) {
        const line = lines[i];
        if (currentOffset + line.length >= offset) {
            return {
                row: row,
                column: offset - currentOffset,
            };
        }
        row++;
        currentOffset += line.length + 1;
    }
    return null;
}
exports.textOffsetToCursorPosition = textOffsetToCursorPosition;
function lineAtRow(body, row) {
    if (!body)
        return '';
    const lines = body.split('\n');
    if (row < 0 || row >= lines.length)
        return '';
    return lines[row];
}
function selectionRangeCurrentLine(selectionRange, body) {
    if (!selectionRange)
        return '';
    return lineAtRow(body, selectionRange.start.row);
}
exports.selectionRangeCurrentLine = selectionRangeCurrentLine;
function selectionRangePreviousLine(selectionRange, body) {
    if (!selectionRange)
        return '';
    return lineAtRow(body, selectionRange.start.row - 1);
}
exports.selectionRangePreviousLine = selectionRangePreviousLine;
function lineLeftSpaces(line) {
    let output = '';
    for (let i = 0; i < line.length; i++) {
        if ([' ', '\t'].indexOf(line[i]) >= 0) {
            output += line[i];
        }
        else {
            break;
        }
    }
    return output;
}
exports.lineLeftSpaces = lineLeftSpaces;
function usePrevious(value) {
    const ref = react_1.useRef();
    react_1.useEffect(() => {
        ref.current = value;
    });
    return ref.current;
}
exports.usePrevious = usePrevious;
function useScrollHandler(editor, webviewRef, onScroll) {
    const editorMaxScrollTop_ = react_1.useRef(0);
    const restoreScrollTop_ = react_1.useRef(null);
    const ignoreNextEditorScrollEvent_ = react_1.useRef(false);
    const scrollTimeoutId_ = react_1.useRef(null);
    // TODO: Below is not needed anymore????
    //
    // this.editorMaxScrollTop_ = 0;
    // // HACK: To go around a bug in Ace editor, we first set the scroll position to 1
    // // and then (in the renderer callback) to the value we actually need. The first
    // // operation helps clear the scroll position cache. See:
    // //
    // this.editorSetScrollTop(1);
    // this.restoreScrollTop_ = 0;
    const editorSetScrollTop = react_1.useCallback((v) => {
        if (!editor)
            return;
        editor.getSession().setScrollTop(v);
    }, [editor]);
    // Complicated but reliable method to get editor content height
    // https://github.com/ajaxorg/ace/issues/2046
    const onAfterEditorRender = react_1.useCallback(() => {
        const r = editor.renderer;
        editorMaxScrollTop_.current = Math.max(0, r.layerConfig.maxHeight - r.$size.scrollerHeight);
        if (restoreScrollTop_.current !== null) {
            editorSetScrollTop(restoreScrollTop_.current);
            restoreScrollTop_.current = null;
        }
    }, [editor, editorSetScrollTop]);
    const scheduleOnScroll = react_1.useCallback((event) => {
        if (scrollTimeoutId_.current) {
            clearTimeout(scrollTimeoutId_.current);
            scrollTimeoutId_.current = null;
        }
        scrollTimeoutId_.current = setTimeout(() => {
            scrollTimeoutId_.current = null;
            onScroll(event);
        }, 10);
    }, [onScroll]);
    const setEditorPercentScroll = react_1.useCallback((p) => {
        ignoreNextEditorScrollEvent_.current = true;
        editorSetScrollTop(p * editorMaxScrollTop_.current);
        scheduleOnScroll({ percent: p });
    }, [editorSetScrollTop, scheduleOnScroll]);
    const setViewerPercentScroll = react_1.useCallback((p) => {
        if (webviewRef.current) {
            webviewRef.current.wrappedInstance.send('setPercentScroll', p);
            scheduleOnScroll({ percent: p });
        }
    }, [scheduleOnScroll]);
    const editor_scroll = react_1.useCallback(() => {
        if (ignoreNextEditorScrollEvent_.current) {
            ignoreNextEditorScrollEvent_.current = false;
            return;
        }
        const m = editorMaxScrollTop_.current;
        const percent = m ? editor.getSession().getScrollTop() / m : 0;
        setViewerPercentScroll(percent);
    }, [editor, setViewerPercentScroll]);
    const resetScroll = react_1.useCallback(() => {
        if (!editor)
            return;
        // Ace Editor caches scroll values, which makes
        // it hard to reset the scroll position, so we
        // need to use this hack.
        // https://github.com/ajaxorg/ace/issues/2195
        editor.session.$scrollTop = -1;
        editor.session.$scrollLeft = -1;
        editor.renderer.scrollTop = -1;
        editor.renderer.scrollLeft = -1;
        editor.renderer.scrollBarV.scrollTop = -1;
        editor.renderer.scrollBarH.scrollLeft = -1;
        editor.session.setScrollTop(0);
        editor.session.setScrollLeft(0);
    }, [editorSetScrollTop, editor]);
    react_1.useEffect(() => {
        if (!editor)
            return () => { };
        editor.renderer.on('afterRender', onAfterEditorRender);
        return () => {
            editor.renderer.off('afterRender', onAfterEditorRender);
        };
    }, [editor]);
    return { resetScroll, setEditorPercentScroll, setViewerPercentScroll, editor_scroll };
}
exports.useScrollHandler = useScrollHandler;
function useRootWidth(dependencies) {
    const { rootRef } = dependencies;
    const [rootWidth, setRootWidth] = react_1.useState(0);
    react_1.useEffect(() => {
        if (!rootRef.current)
            return;
        if (rootWidth !== rootRef.current.offsetWidth)
            setRootWidth(rootRef.current.offsetWidth);
    });
    return rootWidth;
}
exports.useRootWidth = useRootWidth;
//# sourceMappingURL=index.js.map