"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const React = require("react");
const react_1 = require("react");
const CodeMirror = require("codemirror");
require("codemirror/addon/comment/comment");
require("codemirror/addon/dialog/dialog");
require("codemirror/addon/edit/closebrackets");
require("codemirror/addon/edit/continuelist");
require("codemirror/addon/scroll/annotatescrollbar");
require("codemirror/addon/search/matchesonscrollbar");
require("codemirror/addon/search/searchcursor");
const useListIdent_1 = require("./utils/useListIdent");
const useScrollUtils_1 = require("./utils/useScrollUtils");
const useCursorUtils_1 = require("./utils/useCursorUtils");
const useLineSorting_1 = require("./utils/useLineSorting");
const useEditorSearch_1 = require("./utils/useEditorSearch");
const useJoplinMode_1 = require("./utils/useJoplinMode");
require("codemirror/keymap/emacs");
require("codemirror/keymap/vim");
require("codemirror/keymap/sublime"); // Used for swapLineUp and swapLineDown
require("codemirror/mode/meta");
const { shim } = require('lib/shim.js');
const { reg } = require('lib/registry.js');
// Based on http://pypl.github.io/PYPL.html
// +XML (HTML) +CSS and Markdown added
const topLanguages = [
    'python',
    'clike',
    'javascript',
    'jsx',
    'php',
    'r',
    'swift',
    'go',
    'vb',
    'vbscript',
    'ruby',
    'rust',
    'dart',
    'lua',
    'groovy',
    'perl',
    'cobol',
    'julia',
    'haskell',
    'pascal',
    'css',
    'xml',
    'markdown',
];
// Load Top Modes
for (let i = 0; i < topLanguages.length; i++) {
    const mode = topLanguages[i];
    if (CodeMirror.modeInfo.find((m) => m.mode === mode)) {
        require(`codemirror/mode/${mode}/${mode}`);
    }
    else {
        reg.logger().error('Cannot find CodeMirror mode: ', mode);
    }
}
function Editor(props, ref) {
    const [editor, setEditor] = react_1.useState(null);
    const editorParent = react_1.useRef(null);
    // Codemirror plugins add new commands to codemirror (or change it's behavior)
    // This command adds the smartListIndent function which will be bound to tab
    useListIdent_1.default(CodeMirror);
    useScrollUtils_1.default(CodeMirror);
    useCursorUtils_1.default(CodeMirror);
    useLineSorting_1.default(CodeMirror);
    useEditorSearch_1.default(CodeMirror);
    useJoplinMode_1.default(CodeMirror);
    react_1.useImperativeHandle(ref, () => {
        return editor;
    });
    const editor_change = react_1.useCallback((cm, change) => {
        if (props.onChange && change.origin !== 'setValue') {
            props.onChange(cm.getValue());
        }
    }, [props.onChange]);
    // eslint-disable-next-line no-unused-vars, @typescript-eslint/no-unused-vars
    const editor_scroll = react_1.useCallback((_cm) => {
        props.onScroll();
    }, [props.onScroll]);
    // eslint-disable-next-line no-unused-vars, @typescript-eslint/no-unused-vars
    const editor_mousedown = react_1.useCallback((_cm, event) => {
        if (event && event.button === 2) {
            props.onEditorContextMenu();
        }
    }, [props.onEditorContextMenu]);
    // eslint-disable-next-line no-unused-vars, @typescript-eslint/no-unused-vars
    const editor_paste = react_1.useCallback((_cm, _event) => {
        props.onEditorPaste();
    }, [props.onEditorPaste]);
    // eslint-disable-next-line no-unused-vars, @typescript-eslint/no-unused-vars
    const editor_drop = react_1.useCallback((cm, _event) => {
        cm.focus();
    }, []);
    const editor_drag = react_1.useCallback((cm, event) => {
        // This is the type for all drag and drops that are external to codemirror
        // setting the cursor allows us to drop them in the right place
        if (event.dataTransfer.effectAllowed === 'all') {
            const coords = cm.coordsChar({ left: event.x, top: event.y });
            cm.setCursor(coords);
        }
    }, []);
    react_1.useEffect(() => {
        CodeMirror.keyMap.basic = {
            'Left': 'goCharLeft',
            'Right': 'goCharRight',
            'Up': 'goLineUp',
            'Down': 'goLineDown',
            'End': 'goLineEnd',
            'Home': 'goLineStartSmart',
            'PageUp': 'goPageUp',
            'PageDown': 'goPageDown',
            'Delete': 'delCharAfter',
            'Backspace': 'delCharBefore',
            'Shift-Backspace': 'delCharBefore',
            'Tab': 'smartListIndent',
            'Shift-Tab': 'smartListUnindent',
            'Enter': 'insertListElement',
            'Insert': 'toggleOverwrite',
            'Esc': 'singleSelection',
        };
        if (shim.isMac()) {
            CodeMirror.keyMap.default = {
                // MacOS
                'Cmd-A': 'selectAll',
                'Cmd-D': 'deleteLine',
                'Cmd-Z': 'undo',
                'Shift-Cmd-Z': 'redo',
                'Cmd-Y': 'redo',
                'Cmd-Home': 'goDocStart',
                'Cmd-Up': 'goDocStart',
                'Cmd-End': 'goDocEnd',
                'Cmd-Down': 'goDocEnd',
                'Cmd-Left': 'goLineLeft',
                'Cmd-Right': 'goLineRight',
                'Alt-Left': 'goGroupLeft',
                'Alt-Right': 'goGroupRight',
                'Alt-Backspace': 'delGroupBefore',
                'Alt-Delete': 'delGroupAfter',
                'Cmd-[': 'indentLess',
                'Cmd-]': 'indentMore',
                'Cmd-/': 'toggleComment',
                'Cmd-Opt-S': 'sortSelectedLines',
                'Opt-Up': 'swapLineUp',
                'Opt-Down': 'swapLineDown',
                'fallthrough': 'basic',
            };
        }
        else {
            CodeMirror.keyMap.default = {
                // Windows/linux
                'Ctrl-A': 'selectAll',
                'Ctrl-D': 'deleteLine',
                'Ctrl-Z': 'undo',
                'Shift-Ctrl-Z': 'redo',
                'Ctrl-Y': 'redo',
                'Ctrl-Home': 'goDocStart',
                'Ctrl-End': 'goDocEnd',
                'Ctrl-Up': 'goLineUp',
                'Ctrl-Down': 'goLineDown',
                'Ctrl-Left': 'goGroupLeft',
                'Ctrl-Right': 'goGroupRight',
                'Alt-Left': 'goLineStart',
                'Alt-Right': 'goLineEnd',
                'Ctrl-Backspace': 'delGroupBefore',
                'Ctrl-Delete': 'delGroupAfter',
                'Ctrl-[': 'indentLess',
                'Ctrl-]': 'indentMore',
                'Ctrl-/': 'toggleComment',
                'Ctrl-Alt-S': 'sortSelectedLines',
                'Alt-Up': 'swapLineUp',
                'Alt-Down': 'swapLineDown',
                'fallthrough': 'basic',
            };
        }
    }, []);
    react_1.useEffect(() => {
        if (!editorParent.current)
            return () => { };
        const cmOptions = {
            value: props.value,
            screenReaderLabel: props.value,
            theme: props.theme,
            mode: props.mode,
            readOnly: props.readOnly,
            autoCloseBrackets: props.autoMatchBraces,
            inputStyle: 'textarea',
            lineWrapping: true,
            lineNumbers: false,
            indentWithTabs: true,
            indentUnit: 4,
            spellcheck: true,
            allowDropFileTypes: [''],
            keyMap: props.keyMap ? props.keyMap : 'default',
        };
        const cm = CodeMirror(editorParent.current, cmOptions);
        setEditor(cm);
        cm.on('change', editor_change);
        cm.on('scroll', editor_scroll);
        cm.on('mousedown', editor_mousedown);
        cm.on('paste', editor_paste);
        cm.on('drop', editor_drop);
        cm.on('dragover', editor_drag);
        return () => {
            // Clean up codemirror
            cm.off('change', editor_change);
            cm.off('scroll', editor_scroll);
            cm.off('mousedown', editor_mousedown);
            cm.off('paste', editor_paste);
            cm.off('drop', editor_drop);
            cm.off('dragover', editor_drag);
            editorParent.current.removeChild(cm.getWrapperElement());
            setEditor(null);
        };
    }, []);
    react_1.useEffect(() => {
        if (editor) {
            //  Value can also be changed by the editor itself so we need this guard
            //  to prevent loops
            if (props.value !== editor.getValue()) {
                editor.setValue(props.value);
                editor.clearHistory();
            }
            editor.setOption('screenReaderLabel', props.value);
        }
    }, [props.value]);
    react_1.useEffect(() => {
        if (editor) {
            editor.setOption('theme', props.theme);
        }
    }, [props.theme]);
    react_1.useEffect(() => {
        if (editor) {
            editor.setOption('mode', props.mode);
        }
    }, [props.mode]);
    react_1.useEffect(() => {
        if (editor) {
            editor.setOption('readOnly', props.readOnly);
        }
    }, [props.readOnly]);
    react_1.useEffect(() => {
        if (editor) {
            editor.setOption('autoCloseBrackets', props.autoMatchBraces);
        }
    }, [props.autoMatchBraces]);
    react_1.useEffect(() => {
        if (editor) {
            editor.setOption('keyMap', props.keyMap ? props.keyMap : 'default');
        }
    }, [props.keyMap]);
    return React.createElement("div", { style: props.style, ref: editorParent });
}
exports.default = react_1.forwardRef(Editor);
//# sourceMappingURL=Editor.js.map