"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const react_1 = require("react");
const types_1 = require("./types");
const resourceHandling_1 = require("./resourceHandling");
const AsyncActionQueue_1 = require("../../../lib/AsyncActionQueue");
const resourceHandling_2 = require("./resourceHandling");
const { MarkupToHtml } = require('lib/joplin-renderer');
const HtmlToHtml = require('lib/joplin-renderer/HtmlToHtml');
const usePrevious = require('lib/hooks/usePrevious').default;
const Note = require('lib/models/Note');
const Setting = require('lib/models/Setting');
const { reg } = require('lib/registry.js');
const ResourceFetcher = require('lib/services/ResourceFetcher.js');
const DecryptionWorker = require('lib/services/DecryptionWorker.js');
const ResourceEditWatcher = require('lib/services/ResourceEditWatcher/index').default;
function installResourceChangeHandler(onResourceChangeHandler) {
    ResourceFetcher.instance().on('downloadComplete', onResourceChangeHandler);
    ResourceFetcher.instance().on('downloadStarted', onResourceChangeHandler);
    DecryptionWorker.instance().on('resourceDecrypted', onResourceChangeHandler);
    ResourceEditWatcher.instance().on('resourceChange', onResourceChangeHandler);
}
function uninstallResourceChangeHandler(onResourceChangeHandler) {
    ResourceFetcher.instance().off('downloadComplete', onResourceChangeHandler);
    ResourceFetcher.instance().off('downloadStarted', onResourceChangeHandler);
    DecryptionWorker.instance().off('resourceDecrypted', onResourceChangeHandler);
    ResourceEditWatcher.instance().off('resourceChange', onResourceChangeHandler);
}
function resourceInfosChanged(a, b) {
    if (Object.keys(a).length !== Object.keys(b).length)
        return true;
    for (const id in a) {
        const r1 = a[id];
        const r2 = b[id];
        if (!r2)
            return true;
        if (r1.item.updated_time !== r2.item.updated_time)
            return true;
        if (r1.item.encryption_applied !== r2.item.encryption_applied)
            return true;
        if (r1.item.is_shared !== r2.item.is_shared)
            return true;
        if (r1.localState.fetch_status !== r2.localState.fetch_status)
            return true;
    }
    return false;
}
function useFormNote(dependencies) {
    const { syncStarted, noteId, isProvisional, titleInputRef, editorRef, onBeforeLoad, onAfterLoad } = dependencies;
    const [formNote, setFormNote] = react_1.useState(types_1.defaultFormNote());
    const [isNewNote, setIsNewNote] = react_1.useState(false);
    const prevSyncStarted = usePrevious(syncStarted);
    const previousNoteId = usePrevious(formNote.id);
    const [resourceInfos, setResourceInfos] = react_1.useState({});
    function initNoteState(n) {
        return __awaiter(this, void 0, void 0, function* () {
            let originalCss = '';
            if (n.markup_language === MarkupToHtml.MARKUP_LANGUAGE_HTML) {
                const htmlToHtml = new HtmlToHtml();
                const splitted = htmlToHtml.splitHtml(n.body);
                originalCss = splitted.css;
            }
            const newFormNote = {
                id: n.id,
                title: n.title,
                body: n.body,
                is_todo: n.is_todo,
                parent_id: n.parent_id,
                bodyWillChangeId: 0,
                bodyChangeId: 0,
                markup_language: n.markup_language,
                saveActionQueue: new AsyncActionQueue_1.default(300),
                originalCss: originalCss,
                hasChanged: false,
                user_updated_time: n.user_updated_time,
                encryption_applied: n.encryption_applied,
            };
            // Note that for performance reason,the call to setResourceInfos should
            // be first because it loads the resource infos in an async way. If we
            // swap them, the formNote will be updated first and rendered, then the
            // the resources will load, and the note will be re-rendered.
            setResourceInfos(yield resourceHandling_1.attachedResources(n.body));
            setFormNote(newFormNote);
            yield resourceHandling_2.handleResourceDownloadMode(n.body);
            return newFormNote;
        });
    }
    react_1.useEffect(() => {
        // Check that synchronisation has just finished - and
        // if the note has never been changed, we reload it.
        // If the note has already been changed, it's a conflict
        // that's already been handled by the synchronizer.
        if (!prevSyncStarted)
            return () => { };
        if (syncStarted)
            return () => { };
        if (formNote.hasChanged)
            return () => { };
        reg.logger().debug('Sync has finished and note has never been changed - reloading it');
        let cancelled = false;
        const loadNote = () => __awaiter(this, void 0, void 0, function* () {
            const n = yield Note.load(noteId);
            if (cancelled)
                return;
            // Normally should not happened because if the note has been deleted via sync
            // it would not have been loaded in the editor (due to note selection changing
            // on delete)
            if (!n) {
                reg.logger().warn('Trying to reload note that has been deleted:', noteId);
                return;
            }
            yield initNoteState(n);
        });
        loadNote();
        return () => {
            cancelled = true;
        };
    }, [prevSyncStarted, syncStarted, formNote]);
    react_1.useEffect(() => {
        if (!noteId)
            return () => { };
        if (formNote.id === noteId)
            return () => { };
        let cancelled = false;
        reg.logger().debug('Loading existing note', noteId);
        function handleAutoFocus(noteIsTodo) {
            if (!isProvisional)
                return;
            const focusSettingName = noteIsTodo ? 'newTodoFocus' : 'newNoteFocus';
            requestAnimationFrame(() => {
                if (Setting.value(focusSettingName) === 'title') {
                    if (titleInputRef.current)
                        titleInputRef.current.focus();
                }
                else {
                    if (editorRef.current)
                        editorRef.current.execCommand({ name: 'focus' });
                }
            });
        }
        function loadNote() {
            return __awaiter(this, void 0, void 0, function* () {
                const n = yield Note.load(noteId);
                if (cancelled)
                    return;
                if (!n)
                    throw new Error(`Cannot find note with ID: ${noteId}`);
                reg.logger().debug('Loaded note:', n);
                yield onBeforeLoad({ formNote });
                const newFormNote = yield initNoteState(n);
                setIsNewNote(isProvisional);
                yield onAfterLoad({ formNote: newFormNote });
                handleAutoFocus(!!n.is_todo);
            });
        }
        loadNote();
        return () => {
            cancelled = true;
        };
    }, [noteId, isProvisional, formNote]);
    const onResourceChange = react_1.useCallback(function (event = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const resourceIds = yield Note.linkedResourceIds(formNote.body);
            if (!event || resourceIds.indexOf(event.id) >= 0) {
                resourceHandling_1.clearResourceCache();
                setResourceInfos(yield resourceHandling_1.attachedResources(formNote.body));
            }
        });
    }, [formNote.body]);
    react_1.useEffect(() => {
        installResourceChangeHandler(onResourceChange);
        return () => {
            uninstallResourceChangeHandler(onResourceChange);
        };
    }, [onResourceChange]);
    react_1.useEffect(() => {
        if (previousNoteId !== formNote.id) {
            onResourceChange();
        }
    }, [previousNoteId, formNote.id, onResourceChange]);
    react_1.useEffect(() => {
        let cancelled = false;
        function runEffect() {
            return __awaiter(this, void 0, void 0, function* () {
                const r = yield resourceHandling_1.attachedResources(formNote.body);
                if (cancelled)
                    return;
                setResourceInfos((previous) => {
                    return resourceInfosChanged(previous, r) ? r : previous;
                });
            });
        }
        runEffect();
        return () => {
            cancelled = true;
        };
    }, [formNote.body]);
    return { isNewNote, formNote, setFormNote, resourceInfos };
}
exports.default = useFormNote;
//# sourceMappingURL=useFormNote.js.map