"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const react_1 = require("react");
const editorCommandDeclarations_1 = require("../commands/editorCommandDeclarations");
const CommandService_1 = require("../../../lib/services/CommandService");
const { time } = require('lib/time-utils.js');
const BaseModel = require('lib/BaseModel');
const { reg } = require('lib/registry.js');
const { MarkupToHtml } = require('lib/joplin-renderer');
const commandsWithDependencies = [
    require('../commands/showLocalSearch'),
    require('../commands/focusElementNoteTitle'),
    require('../commands/focusElementNoteBody'),
];
function editorCommandRuntime(declaration, editorRef) {
    return {
        execute: (props) => {
            console.info('Running editor command:', declaration.name, props);
            if (!editorRef.current.execCommand) {
                reg.logger().warn('Received command, but editor cannot execute commands', declaration.name);
            }
            else {
                const execArgs = {
                    name: declaration.name,
                    value: props.value,
                };
                if (declaration.name === 'insertDateTime') {
                    execArgs.name = 'insertText';
                    execArgs.value = time.formatMsToLocal(new Date().getTime());
                }
                editorRef.current.execCommand(execArgs);
            }
        },
        isEnabled: (props) => {
            if (props.markdownEditorViewerOnly)
                return false;
            if (!props.noteId)
                return false;
            const note = BaseModel.byId(props.notes, props.noteId);
            if (!note)
                return false;
            return note.markup_language === MarkupToHtml.MARKUP_LANGUAGE_MARKDOWN;
        },
        mapStateToProps: (state) => {
            return {
                // True when the Markdown editor is active, and only the viewer pane is visible
                // In this case, all editor-related shortcuts are disabled.
                markdownEditorViewerOnly: state.settings['editor.codeView'] && state.noteVisiblePanes.length === 1 && state.noteVisiblePanes[0] === 'viewer',
                noteVisiblePanes: state.noteVisiblePanes,
                notes: state.notes,
                noteId: state.selectedNoteIds.length === 1 ? state.selectedNoteIds[0] : null,
            };
        },
    };
}
function useWindowCommandHandler(dependencies) {
    const { setShowLocalSearch, noteSearchBarRef, editorRef, titleInputRef } = dependencies;
    react_1.useEffect(() => {
        for (const declaration of editorCommandDeclarations_1.default) {
            CommandService_1.default.instance().registerRuntime(declaration.name, editorCommandRuntime(declaration, editorRef));
        }
        const dependencies = {
            editorRef,
            setShowLocalSearch,
            noteSearchBarRef,
            titleInputRef,
        };
        for (const command of commandsWithDependencies) {
            CommandService_1.default.instance().registerRuntime(command.declaration.name, command.runtime(dependencies));
        }
        return () => {
            for (const declaration of editorCommandDeclarations_1.default) {
                CommandService_1.default.instance().unregisterRuntime(declaration.name);
            }
            for (const command of commandsWithDependencies) {
                CommandService_1.default.instance().unregisterRuntime(command.declaration.name);
            }
        };
    }, [editorRef, setShowLocalSearch, noteSearchBarRef, titleInputRef]);
}
exports.default = useWindowCommandHandler;
//# sourceMappingURL=useWindowCommandHandler.js.map